"""CSV lccr reader"""

"""
Copyright (c) 2020, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os

from collections import OrderedDict
import logging
logger = logging.getLogger(__name__)
import numpy


def process(config=None,
            filename=None,
            cellname=None,
            expname=None,
            stim_feats=None,
            idx_file=None,
            ljp=0, v_corr=0):
    """Read recordings from an csv file"""
    path = config['path']
    cells = config['cells']

    data = OrderedDict()
    data['voltage'] = []
    data['current'] = []
    data['dt'] = []

    data['t'] = []
    data['ton'] = []
    data['toff'] = []
    data['tend'] = []
    data['amp'] = []
    data['hypamp'] = []
    data['filename'] = []

    fln = os.path.join(path, cellname, filename + '.txt')
    if not os.path.exists(fln):
        raise Exception(
            'Please provide a string with filename of csv file, '
            'current path not found: %s',
            fln)

    exp_options = cells[cellname]['experiments'][expname]

    if (('dt' not in exp_options) or
        ('amplitudes' not in exp_options) or
        ('hypamp' not in exp_options) or
            ('ton' not in exp_options) or ('toff' not in exp_options)):
        raise Exception('Please provide additional options for LCCR csv')

    dt = exp_options['dt']
    hypamp = exp_options['hypamp']
    ton = exp_options['ton']
    toff = exp_options['toff']
    amplitudes = exp_options['amplitudes']

    import csv
    with open(fln, 'rt') as f:
        reader = csv.reader(f, delimiter='\t')
        columns = list(zip(*reader))
        length = numpy.shape(columns)[1]

        for ic, column in enumerate(columns):

            voltage = numpy.zeros(length)
            for istr, string in enumerate(column):
                if (string != "-") and (string != ""):
                    voltage[istr] = float(string)

            t = numpy.arange(len(voltage)) * dt
            amp = amplitudes[ic]
            voltage = voltage - ljp  # correct liquid junction potential

            # remove last 100 ms
            voltage = voltage[0:int(-100. / dt)]
            t = t[0:int(-100. / dt)]
            ion = int(ton / dt)
            ioff = int(toff / dt)
            current = []
            current = numpy.zeros(len(voltage))
            current[ion:ioff] = amp

            if ('exclude' in cells[cellname] and
                any(abs(cells[cellname]['exclude'][idx_file] - amp) <
                    1e-4)):

                logger.info(" Not using trace with amplitude %f", amp)

            else:
                data['voltage'].append(voltage)
                data['current'].append(current)
                data['t'].append(t)

                data['dt'].append(numpy.float(dt))
                data['ton'].append(numpy.float(ton))
                data['toff'].append(numpy.float(toff))
                data['amp'].append(numpy.float(amp))
                data['hypamp'].append(numpy.float(hypamp))
                data['filename'].append(filename)

    return data



"""Tool functions"""

"""
Copyright (c) 2020, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import os
import json

import numpy


def makedir(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


def makedirs(filename):  # also accepts filename
    if not os.path.exists(os.path.dirname(filename)):
        os.makedirs(os.path.dirname(filename))


def print_dict(v, prefix=''):
    if isinstance(v, dict):
        for k, v2 in v.items():
            p2 = "{}['{}']".format(prefix, k)
            print_dict(v2, p2)
    elif isinstance(v, list):
        for i, v2 in enumerate(v):
            p2 = "{}[{}]".format(prefix, i)
            print_dict(v2, p2)
    else:
        print('{}: {} {}'.format(prefix, numpy.shape(v), type(v)))


def collapse_json(text, indent=12):
    """Compacts a string of json data by collapsing whitespace after the
    specified indent level

    NOTE: will not produce correct results when indent level is not a multiple
    of the json indent level
    """
    initial = " " * indent
    out = []  # final json output
    sublevel = []  # accumulation list for sublevel entries
    pending = None  # holder for consecutive entries at exact indent level
    for line in text.splitlines():
        if line.startswith(initial):
            if line[indent] == " ":
                # found a line indented further than the indent level, so add
                # it to the sublevel list
                if pending:
                    # the first item in the sublevel will be the pending item
                    # that was the previous line in the json
                    sublevel.append(pending)
                    pending = None
                item = line.strip()
                sublevel.append(item)
                if item.endswith(","):
                    sublevel.append(" ")
            elif sublevel:
                # found a line at the exact indent level *and* we have sublevel
                # items. This means the sublevel items have come to an end
                sublevel.append(line.strip())
                out.append("".join(sublevel))
                sublevel = []
            else:
                # found a line at the exact indent level but no items indented
                # further, so possibly start a new sub-level
                if pending:
                    # if there is already a pending item, it means that
                    # consecutive entries in the json had the exact same
                    # indentation and that last pending item was not the start
                    # of a new sublevel.
                    out.append(pending)
                pending = line.rstrip()
        else:
            if pending:
                # it's possible that an item will be pending but not added to
                # the output yet, so make sure it's not forgotten.
                out.append(pending)
                pending = None
            if sublevel:
                out.append("".join(sublevel))
            out.append(line)
    return "\n".join(out)


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (numpy.int_, numpy.intc, numpy.intp, numpy.int8,
                            numpy.int16, numpy.int32, numpy.int64, numpy.uint8,
                            numpy.uint16, numpy.uint32, numpy.uint64)):
            return int(obj)
        elif isinstance(obj, (numpy.float_, numpy.float16, numpy.float32,
                              numpy.float64)):
            return float(obj)
        elif isinstance(obj, numpy.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)



"""Extra e-features functions"""

"""
Copyright (c) 2020, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import numpy
import matplotlib.pyplot as plt


def spikerate_tau_jj(peaktimes):
    """Spike rate tau feature"""

    peaktimes = numpy.array(peaktimes)  # keep ms
    freq = 1. / numpy.diff(peaktimes)
    times = peaktimes[:-1]
    times = times - times[0]

    n = len(times)
    S = [0]
    for k in range(1, n):
        S.append(
            S[k - 1] + 1 / 2. * (freq[k] + freq[k - 1]) *
            (times[k] - times[k - 1]))
    S = numpy.array(S)

    M1 = sum((times - times[0])**2)
    M2 = sum((times - times[0]) * S)
    M3 = M2
    M4 = sum(S**2)

    V1 = sum((freq - freq[0]) * (times - times[0]))
    V2 = sum((freq - freq[0]) * S)

    invdet = 1 / (M1 * M4 - M3 * M2)
    I2 = invdet * -M2
    I4 = invdet * M1
    B1 = I2 * V1 + I4 * V2
    slope = B1
    tau = abs(-1. / slope)

    if tau > 1000.:
        tau = None

    return tau


def spikerate_drop(peaktimes, nstart=1, nend=2):
    """Spike rate drop feature"""

    peaktimes = numpy.array(peaktimes) * 1e-3  # convert from ms to s
    freq = 1. / numpy.diff(peaktimes)

    drop = numpy.mean(freq[-nend:]) / numpy.mean(freq[:nstart])

    return min(drop, 1.)


def spikerate_tau_log(peaktimes):
    """Spike rate tau log feature"""

    peaktimes = numpy.array(peaktimes)  # keep ms
    freq = 1. / numpy.diff(peaktimes)

    npeaks = len(peaktimes)
    used_isi = int(numpy.ceil(npeaks / 1.))

    # assume last frequency is f_inf (adapted frequency)
    # normalise by this value but do not include
    steady = freq[-1]

    freq = abs(freq - steady)[:used_isi + 1]
    times = peaktimes[0:used_isi + 1]
    log_freq = numpy.log(freq)

    i_use = []
    for i, f in enumerate(log_freq):
        if numpy.isfinite(f):
            i_use.append(i)
        else:
            break

    times = times[i_use]
    log_freq = log_freq[i_use]

    i_use = [0]
    for i, f in enumerate(numpy.diff(log_freq)):
        if f < 0:
            i_use.append(i + 1)
        else:
            break

    times = times[i_use]
    log_freq = log_freq[i_use]

    slope, _ = numpy.polyfit(times, log_freq, 1)

    tau = abs(-1. / slope)

    if tau > 1000.:
        tau = None

    return tau


def spikerate_slope(peaktimes):
    """Spike rate slope feature"""

    peaktimes = numpy.array(peaktimes) * 1e-3
    isi = numpy.diff(peaktimes)
    isi = isi
    x = numpy.arange(1, len(isi) + 1)

    times = peaktimes[:-1]
    log_isi = numpy.log(isi)
    log_x = numpy.log(x)
    log_times = numpy.log(times)
    freq = 1. / isi
    freq = freq - freq[-1]

    plt.subplot(5, 1, 1)
    plt.plot(times, isi, '*')
    plt.subplot(5, 1, 2)
    plt.plot(times, log_isi, '*')
    plt.subplot(5, 1, 3)
    plt.plot(log_times, log_isi, '*')
    plt.subplot(5, 1, 4)
    plt.plot(x, log_isi, '*')
    plt.subplot(5, 1, 5)
    plt.plot(log_x, log_isi, '*')
    plt.show()

    slope_log, _ = numpy.polyfit(log_x, log_isi, 1)
    slope_semilog, _ = numpy.polyfit(x, log_isi, 1)

    print(slope_log, slope_semilog)

    return slope_log


def spikerate_tau_slope(peaktimes):
    """Spike rate tau slope feature"""

    peaktimes = numpy.array(peaktimes) * 1e-3
    freq = 1. / numpy.diff(peaktimes)

    npeaks = len(peaktimes)
    used_isi = int(numpy.ceil(npeaks / 1.))

    # assume last frequency is f_inf (adapted frequency)
    # normalise by this value but do not include
    steady = freq[-1]

    freq = abs(freq - steady)[:used_isi + 1]
    times = peaktimes[0:used_isi + 1]
    log_freq = numpy.log(freq)

    i_use = []
    for i, f in enumerate(log_freq):
        if numpy.isfinite(f):
            i_use.append(i)
        else:
            break

    times = times[i_use]
    log_freq = log_freq[i_use]

    # i_use = [0]
    # for i, f in enumerate(numpy.diff(log_freq)):
    #     if f<0:
    #         i_use.append(i+1)
    #     else:
    #         break
    #
    # times = times[i_use]
    # log_freq = log_freq[i_use]

    plt.subplot(3, 1, 1)
    plt.plot(times, log_freq, '*')
    plt.subplot(3, 1, 2)
    plt.plot(times[1:], numpy.diff(log_freq), '*')
    plt.subplot(3, 1, 3)
    plt.plot(times[1:], numpy.diff(log_freq) / numpy.diff(times), '*')
    plt.show()

    tau = -numpy.mean(numpy.diff(log_freq) / numpy.diff(times))
    if numpy.isnan(tau):
        tau = 0

    return tau


def spikerate_tau_fit(peaktimes, nstart=1, nend=2):
    """Spike rate tau fit feature"""

    from scipy.optimize import curve_fit
    from functools import partial

    def func(x, tau, a, d):
        return a * numpy.exp(-x / tau) + d

    peaktimes = numpy.array(peaktimes)  # keep ms
    freq = 1. / numpy.diff(peaktimes)
    times = peaktimes[:-1]
    times = times - times[0]

    d = freq[-nend]
    a = freq[:nstart] - d

    func_ = partial(func, a=a, d=d)

    popt, pcov = curve_fit(func_, times, freq, p0=(1e9))

    tau = abs(popt[0])

    if tau > 1000.:
        tau = None

    return tau


def main():
    """Main"""

    tau = 200e-3
    fstart = 5.
    baseline = 10.
    tmax = 2.
    drop = baseline / (fstart + baseline)

    numpy.random.seed(1)
    peaktime = 1. / (fstart + baseline)
    peaktimes = [0]
    while peaktime < tmax:
        peaktimes.append(peaktime)
        next_freq = numpy.exp(-peaktime / tau) * fstart + baseline
        r = numpy.random.normal(0, 1e-9 * baseline, 1)[0]
        peaktime += 1. / (next_freq + r)
    peaktimes = numpy.array(peaktimes) + 200.

    print("target tau:", tau * 1e3)

    tau = spikerate_tau_jj(peaktimes * 1e3)  # convert from s to ms
    print("jj:", tau)

    tau = spikerate_tau_log(peaktimes * 1e3)  # convert from s to ms
    print("log:", tau)

    tau = spikerate_tau_fit(peaktimes * 1e3)  # convert from s to ms
    print("fit:", tau)

    tau = spikerate_slope(peaktimes * 1e3)  # convert from s to ms
    print("slope:", tau)

    print("target drop:", drop)
    drop = spikerate_drop(peaktimes * 1e3)  # convert from s to ms
    print(drop)

    f = 1. / numpy.diff(peaktimes)
    times = peaktimes[:-1]

    plt.figure()
    plt.plot(times, f, 'ko-')
    plt.show()


if __name__ == '__main__':
    main()



"""Igor reader for binary wave format"""

"""
Copyright (c) 2020, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import re
import numpy as np
from igor import binarywave

# as we support Python 2.7/3, we have a special import for StringIO
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


def _bytes_to_str(bytes_):
    """concatenates an iterable of bytes to a str

    Args:
        bytes_(numpy.ndarray): array of bytes

    Returns:
        str of concatenated bytes with omitted `x00` bytes
    """
    return ''.join(bytes_[bytes_ != b''].astype(str))


class IgorHeader(object):
    """Header metaclass information.
    """

    def __init__(self, version, content):
        header = content['wave_header']
        self.bname = header['bname'].decode('utf-8')
        self.dUnits = _bytes_to_str(header['dataUnits'])
        self.npnts = header['npnts']
        self.wavenotes = content['note'].decode('utf-8')

        if version == 5:
            self.xUnits = _bytes_to_str(header['dimUnits'])
            self.dx = header['sfA'][0]
            self.next = header['next']
            self.creationDate = header['creationDate']
            self.modDate = header['modDate']
            self.sfA = header['sfA'].astype(np.double)
            self.dimUnits = header['dimUnits'].astype(str)
            self.fsValid = header['fsValid']
            self.topFullScale = header['topFullScale']
            self.botFullScale = header['botFullScale']
            self.dataEUnits = header['dataEUnits']
            self.dimEUnits = header['dimEUnits']
            self.dimLabels = header['dimLabels']
            self.waveNoteH = header['waveNoteH']
            # self.platform = header['platform']
        else:
            self.xUnits = _bytes_to_str(header['xUnits'])
            self.dx = header['hsA']


def read_wave_notes(wavenotes):
    """parse wavenotes to collect them into a dict

    Args:
        wavenotes(str): string of wavenotes

    Returns:
        dict of wavenotes
    """
    wavenotes = dict(re.findall("(.+?):(.+?);", wavenotes))
    return wavenotes


def read_from_binary(content):
    """Reads Igor's (Wavemetric) binary wave format represented as
    `content` string. Basically it applies `read_from_handle` to `content`
    wrapped into a file handler.

    Args:
        content(str): string

    Returns:
        see `read_from_handle` output
    """
    file_handle = StringIO(content)
    return read_from_handle(file_handle)


def read(filename):
    """Reads Igor's (Wavemetric) binary wave format from a file under
    `filename` path.

    Args:
        filename(str):

    Returns:
        see `read_from_handle` output
    """

    with open(filename, "rb") as f:
        return read_from_handle(f)


def read_from_handle(f):
    """Reads Igor's (Wavemetric) binary wave format, .ibw or .bwav, files.

    Args:
        f(file): file handle

    Returns:
        A tuple of (headerType instance, numpy vector) where `headerType
        instance` contains a meta info about the wave and `numpy vector`
        contains wave data. `numpy vector` is writeable.
    """

    data = binarywave.load(f)
    version = data['version']
    assert version in (2, 3, 5), "Fileversion is '" + \
                                 str(version) + "', not supported"

    content = data['wave']
    wdata = np.copy(content['wData'])
    header = IgorHeader(version, content)
    return header, wdata



"""Axon reader"""

"""
Copyright (c) 2020, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from neo import io
from collections import OrderedDict
import logging
logger = logging.getLogger(__name__)
import numpy
import os
import pprint
import neo.rawio.axonrawio


def process(config=None,
            filename=None,
            cellname=None,
            expname=None,
            stim_feats=None,
            idx_file=None,
            ljp=0, v_corr=0):
    """Read recordings from an axon file"""
    path = config['path']
    cells = config['cells']
    # features = config['features']
    # options = config['options']

    data = OrderedDict()
    data['voltage'] = []
    data['current'] = []
    data['dt'] = []

    data['t'] = []
    data['ton'] = []
    data['toff'] = []
    data['tend'] = []
    data['amp'] = []
    data['hypamp'] = []
    data['filename'] = []

    logger.debug(" Adding axon file %s", filename)

    f = os.path.join(path, cellname, filename + '.abf')
    r = io.AxonIO(filename=f)

    # read header

    # Below line doesn't work anymore due to api change
    # Now using rawio
    # header = r.read_header()

    header = neo.rawio.axonrawio.parse_axon_soup(f)

    # read sampling rate
    sampling_rate = 1.e6 / header['protocol']['fADCSequenceInterval']

    dt = 1. / int(sampling_rate) * 1e3
    # version = header['fFileVersionNumber']  # read file version
    bl = r.read_block(lazy=False)

    stim_info = None
    if 'stim_info' in cells[cellname]['experiments'][expname]:
        stim_info = cells[cellname]['experiments'][expname]['stim_info']

    else:

        # read stimulus features if present
        stim_feats = []
        if 'stim_feats' in cells[cellname]['experiments'][expname]:
            stim_feats = cells[cellname]['experiments'][expname]['stim_feats']

        all_stims = []
        if stim_feats:
            res = stim_feats_from_meta(stim_feats, len(bl.segments), idx_file)
            if res[0]:
                all_stims = res[1]
            else:
                print(res[1])
        if not all_stims:
            res = stim_feats_from_header(header)
            if res[0]:
                all_stims = res[1]
            else:
                pprint.pprint(
                    "No valid stimulus was found in metadata or files. \
                        Skipping current file")
                return

    # for all segments in file
    for i_seg, seg in enumerate(bl.segments):

        # dt = 1./int(seg.analogsignals[0].sampling_rate) * 1e3

        if stim_info is not None:

            voltage = numpy.array(
                seg.analogsignals[0]).astype(
                numpy.float64).flatten()
            current = numpy.array(
                seg.analogsignals[1]).astype(
                numpy.float64).flatten()
            t = numpy.arange(len(voltage)) * dt

            ton = stim_info['ton']
            toff = stim_info['toff']
            ion = int(ton / dt)
            ioff = int(toff / dt)

            if 'tamp' in stim_info:
                tamp = [int(stim_info['tamp'][0] / dt),
                        int(stim_info['tamp'][1] / dt)]
            else:
                tamp = [ion, ioff]

            i_unit = stim_info['i_unit']

            if i_unit == 'A':
                current = current * 1e9  # nA
            elif i_unit == 'pA':
                current = current * 1e-3  # nA
            else:
                raise Exception(
                    "Unit current not configured!")

            amp = numpy.nanmean(current[tamp[0]:tamp[1]])
            hypamp = numpy.nanmean(current[0:ion])

        else:

            voltage = numpy.array(seg.analogsignals[0]).astype(numpy.float64)
            t = numpy.arange(len(voltage)) * dt

            ton = all_stims[i_seg][1]
            toff = all_stims[i_seg][2]
            amp = numpy.float64(all_stims[i_seg][3])

            ion = int(ton / dt)
            ioff = int(toff / dt)

            current = []
            current = numpy.zeros(len(voltage))
            current[ion:ioff] = amp

            # estimate hyperpolarization current
            hypamp = numpy.mean(current[0:ion])

            # clean voltage from transients
            voltage[ion:ion + int(numpy.ceil(0.4 / dt))] = \
                voltage[ion + int(numpy.ceil(0.4 / dt))]
            voltage[ioff:ioff + int(numpy.ceil(0.4 / dt))] = \
                voltage[ioff + int(numpy.ceil(0.4 / dt))]

        # normalize membrane potential to known value (given in UCL excel
        # sheet)
        if v_corr:
            if len(v_corr) == 1 and v_corr[0] != 0.0:
                voltage = voltage - numpy.mean(voltage[0:ion]) + v_corr[0]
            elif len(v_corr) - 1 >= idx_file and v_corr[idx_file] != 0.0:
                voltage = voltage - numpy.mean(voltage[0:ion]) \
                    + v_corr[idx_file]

        voltage = voltage - ljp

        # clip spikes after stimulus so they are not analysed
        voltage[ioff:] = numpy.clip(voltage[ioff:], -300, -40)

        if ('exclude' in cells[cellname] and
                any(abs(cells[cellname]['exclude'][idx_file] - amp) < 1e-4)):
            continue  # llb

        else:
            data['voltage'].append(voltage)
            data['current'].append(current)
            data['dt'].append(dt)

            data['t'].append(t)
            data['tend'].append(t[-1])
            data['ton'].append(ton)
            data['toff'].append(toff)
            data['amp'].append(amp)
            data['hypamp'].append(hypamp)
            data['filename'].append(filename)

    return data


def stim_feats_from_meta(stim_feats, num_segments, idx_file):
    """Get stimulus features from metadata (author: Luca Leonardo
        Bologna)"""
    if not stim_feats:
        return (0, "Empty metadata in file")
    elif len(stim_feats) - 1 < idx_file and len(stim_feats) != 1:
        return (0, "Stimulus dictionaries are different \
                from the number of files")
    else:
        # array for storing all stimulus features
        all_stim_feats = []

        # for every segment in the axon file
        for i in range(num_segments):

            # read current stimulus dict
            if len(stim_feats) == 1:
                crr_dict = stim_feats[0]
            else:
                crr_dict = stim_feats[idx_file]

            # read stimulus information
            ty = str(crr_dict['stimulus_type'])
            tu = crr_dict['stimulus_time_unit']
            st = crr_dict['stimulus_start']
            en = crr_dict['stimulus_end']
            u = str(crr_dict['stimulus_unit'])
            fa = float(format(crr_dict['stimulus_first_amplitude'], '.3f'))
            inc = float(format(crr_dict['stimulus_increment'], '.3f'))
            if tu == 's':
                st = st * 1e3
                en = en * 1e3
            # compute current stimulus amplitude
            crr_val = float(format(fa + inc * float(format(i, '.3f')), '.3f'))
            crr_stim_feats = (ty, st, en, crr_val, u)

            # store current tuple
            all_stim_feats.append(crr_stim_feats)
        return (1, all_stim_feats)


def stim_feats_from_header(header):
    """Get stimulus features from the file header (author: Luca Leonardo
    Bologna)"""
    # read sampling rate
    sampling_rate = 1.e6 / header['protocol']['fADCSequenceInterval']

    # read file version
    version = header['fFileVersionNumber']

    # extract protocol for version >=.2
    if version >= 2.:
        # prot = r.read_protocol() # read protocol
        # read info for DAC
        dictEpochInfoPerDAC = header['dictEpochInfoPerDAC']

        # if field is empty
        if not (dictEpochInfoPerDAC):
            return (0, "No 'dictEpochInfoPerDAC' field")

        # if field is not empty, read all stimulus segments
        else:
            valid_epoch_dicts = [k for k, v in dictEpochInfoPerDAC.items()
                                 if bool(v)]

            # if more than one channel is activated for the stimulus
            # or a number of epochs different than 3 is found
            if len(valid_epoch_dicts) != 1 or len(dictEpochInfoPerDAC[0]) != 3:
                return (0, 'Exiting. More than one channel used \
                        for stimulation')
            else:
                # read all stimulus epochs

                # TODO IS THIS CORRECT ?
                k = valid_epoch_dicts[-1]
                stim_epochs = dictEpochInfoPerDAC[k]
                # read enabled waveforms
                stim_ch_info = [(i['DACChNames'], i['DACChUnits'],
                                 i['nDACNum']) for i in header['listDACInfo']
                                if bool(i['nWaveformEnable'])]

                # if epoch initial levels and increment are not
                # compatible with a step stimulus
                if (stim_epochs[0]['fEpochInitLevel'] !=
                        stim_epochs[2]['fEpochInitLevel'] or
                    stim_epochs[0]['fEpochLevelInc'] !=
                    stim_epochs[2]['fEpochLevelInc'] or
                    float(format(stim_epochs[0]['fEpochLevelInc'], '.3f')) != 0
                    or (len(stim_ch_info) != 1 or
                        stim_ch_info[0][2] != k)):
                    # return 0 with message
                    return (0, "A stimulus different from the steps \
                                has been detected")
                else:
                    ty = "step"
                    u = stim_ch_info[0][1]
                    # number of ADC channels
                    nADC = header['sections']['ADCSection']['llNumEntries']
                    # number of samples per episode
                    nSam = header['protocol']['lNumSamplesPerEpisode'] / nADC
                    # number of actual episodes
                    nEpi = header['lActualEpisodes']

                    # read first stimulus epoch
                    e_zero = header['dictEpochInfoPerDAC'][
                        stim_ch_info[0][2]][0]
                    # read second stimulus epoch
                    e_one = header['dictEpochInfoPerDAC'][
                        stim_ch_info[0][2]][1]
                    # index of stimulus beginning
                    i_last = int(nSam * 15625 / 10**6)
                    # create array for all stimulus info
                    all_stim_info = []

                    # step increment
                    e_one_inc = float(format(e_one['fEpochLevelInc'],
                                             '.3f'))
                    # step initial level
                    e_one_init_level = float(format(e_one['fEpochInitLevel'],
                                                    '.3f'))

                    # for every episode, compute stimulus start, stimulus end,
                    # stimulus value
                    for epiNum in range(nEpi):
                        st = i_last + e_zero['lEpochInitDuration'] + \
                            e_zero['lEpochDurationInc'] * epiNum
                        en = st + e_one['lEpochInitDuration'] +  \
                            e_one['lEpochDurationInc'] * epiNum
                        crr_val_full = float(format(e_one_init_level +
                                                    e_one_inc * epiNum, '.3f'))
                        crr_val = float(format(crr_val_full, '.3f'))
                        st = 1 / sampling_rate * st * 1e3
                        en = 1 / sampling_rate * en * 1e3
                        all_stim_info.append((ty, st, en, crr_val, u))
                    return (1, all_stim_info)



"""BluePyEfe setup """

"""
Copyright (c) 2020, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import setuptools
import versioneer

setuptools.setup(
    name='bluepyefe',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Blue Brain Python E-feature extraction',
    long_description="The Blue Brain Python E-feature extraction Library ("
    "BluePyEfe) aims at easing the process of reading experimental recordings "
    "and extracting batches of electrical features from them. To do so, it "
    "combines trace reading functions and features extraction functions from "
    "the eFel library. BluePyEfe outputs protocols and features files in a "
    "format that can then be used by BluePyOpt for electrical model building "
    "purposes.",
    packages=setuptools.find_packages(),
    author="BlueBrain Project, EPFL",
    license="LGPLv3",
    keywords=(
        'neuroscience',
        'BlueBrainProject'),
    url='https://github.com/BlueBrain/BluePyEfe',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'License :: OSI Approved :: GNU Lesser General Public '
        'License v3 (LGPLv3)',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Operating System :: POSIX',
        'Topic :: Scientific/Engineering',
        'Topic :: Utilities'],
    install_requires=[
        'igor',
        'neo',
        'matplotlib',
        'efel',
        'sh',
        'pandas',
        'scipy'],
)
